# -*- coding: utf-8 -*-
"""Data models functions"""
import os
import logging
import pickle

logger = logging.getLogger(__name__)


class SavePoint(object):
    """Model: SavePoint"""
    default_file_name = 'savepoint.pickle'

    def __init__(self, db_dir):
        self.db_file_path = os.path.join(db_dir, self.default_file_name)
        if os.path.exists(self.db_file_path):
            with open(self.db_file_path, 'rb') as fp:
                self.data = pickle.load(fp)
        else:
            self.data = {}

    def set_last_datetime(self, datetime):
        self.data['last_datetime'] = datetime

    def get_last_datetime(self):
        return self.data.get('last_datetime')

    def save(self):
        logger.info('SavePoint value change to %s' % self.get_last_datetime())
        with open(self.db_file_path, 'wb') as fp:
            pickle.dump(self.data, fp)


class RequestsData(object):
    """Model: RequestsData"""

    def __init__(self, date, db_dir):
        self.date = date
        self.db_file_path = os.path.join(db_dir, '%s.pickle' % date)
        if os.path.exists(self.db_file_path):
            with open(self.db_file_path, 'rb') as fp:
                self.data = pickle.load(fp)
        else:
            self.data = {}

    def save(self):
        with open(self.db_file_path, 'wb') as fp:
            pickle.dump(self.data, fp)


# Utils for requests data

def merge_urls_data_to(to, food={}):
    """Merge urls data"""
    if not to:
        to.update(food)

    for url, data in food.items():
        if url not in to:
            to[url] = data
        else:
            to[url] = to[url].merge_with(data)


def merge_requests_data_to(to, food={}):
    """Merge a small analyzed result to a big one, this function will modify the 
    original ``to``"""
    if not to:
        to.update(food)

    to['requests_counter']['normal'] += food['requests_counter']['normal']
    to['requests_counter']['slow'] += food['requests_counter']['slow']
    to['total_slow_duration'] += food['total_slow_duration']

    for group_name, urls in food['data_details'].items():
        if group_name not in to['data_details']:
            to['data_details'][group_name] = urls
        else:
            to_urls = to['data_details'][group_name]
            to_urls['duration_agr_data'] = to_urls['duration_agr_data'].merge_with(
                    urls['duration_agr_data'])

            # Merge urls data
            merge_urls_data_to(to_urls['urls'], urls['urls'])




# -*- coding: utf-8 -*-
from __future__ import print_function
import pkg_resources
from uwsgi_sloth.utils import force_text


def echo_conf(args):
    print(force_text(pkg_resources.resource_string('uwsgi_sloth', "sample.conf")))


def load_subcommand(subparsers):
    """Load this subcommand"""
    parser_analyze = subparsers.add_parser('echo_conf', help='Echo sample configuration file')
    parser_analyze.set_defaults(func=echo_conf)
    




# -*- coding: utf-8 -*-
"""Template shortcut & filters"""
import os
import datetime
from jinja2 import Environment, FileSystemLoader

from uwsgi_sloth.settings import ROOT
from uwsgi_sloth import settings, __VERSION__

template_path = os.path.join(ROOT, 'templates')
env = Environment(loader=FileSystemLoader(template_path))

# Template filters

def friendly_time(msecs):
    secs, msecs = divmod(msecs, 1000)
    mins, secs = divmod(secs, 60)
    hours, mins = divmod(mins, 60)
    if hours:
        return '%dh%dm%ds' % (hours, mins, secs)
    elif mins:
        return '%dm%ds' % (mins, secs)
    elif secs:
        return '%ds%dms' % (secs, msecs)
    else:
        return '%.2fms' % msecs

env.filters['friendly_time'] = friendly_time


def render_template(template_name, context={}):
    template = env.get_template(template_name)
    context.update(
        SETTINGS=settings,
        now=datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
        version='.'.join(map(str, __VERSION__)))
    return template.render(**context)





# -*- coding: utf-8 -*-
import os

ROOT = os.path.dirname(os.path.abspath(__file__))

FILTER_METHODS = ('GET', 'POST', 'PUT', 'DELETE', 'PATCH')
FILTER_STATUS = ('200', )

LIMIT_URL_GROUPS = 200
LIMIT_PER_URL_GROUP = 20
DEFAULT_MIN_MSECS = 200

STATIC_PATH_BOOTSTRAP = 'https://maxcdn.bootstrapcdn.com/bootstrap/3.1.1/css/bootstrap.min.css'
STATIC_PATH_JQUERY = 'https://code.jquery.com/jquery-1.7.2.min.js'

STATIC_PATH_SORTABLE_CSS = 'https://cdnjs.cloudflare.com/ajax/libs/sortable/0.8.0/css/sortable-theme-bootstrap.min.css'
STATIC_PATH_SORTABLE = 'https://cdnjs.cloudflare.com/ajax/libs/sortable/0.8.0/js/sortable.min.js'

REALTIME_UPDATE_INTERVAL = 5 * 60



# -*- coding: utf-8 -*-
from uwsgi_sloth.structures import ValuesAggregation

def test_ValuesAggregation():
    agr = ValuesAggregation()
    agr.add_values(range(1, 101))
    assert agr.get_result() == {'max': 100, 'avg': 50.5, 'min': 1}
    assert agr.avg == 50.5

    # Test merge with
    agr1 = ValuesAggregation(values=range(1, 11))
    agr2 = ValuesAggregation(values=range(-10, 0))
    agr3 = ValuesAggregation(values=range(100, 201))
    assert agr1.merge_with(agr2).get_result() == {'max': 10, 'avg': 0.0, 'min': -10}
    assert agr1.merge_with(agr3).get_result() == {
        'max': 200, 
        'avg': (sum(range(1, 11)) + sum(range(100, 201))) / 111.0, 
        'min': 1
    }




# -*- coding: utf-8 -*-
"""Analyzer for uwsgi log"""
import re
import copy
import datetime
from uwsgi_sloth.utils import total_seconds
from uwsgi_sloth.structures import ValuesAggregation
from uwsgi_sloth.settings import FILTER_METHODS, FILTER_STATUS, LIMIT_URL_GROUPS, \
                                 LIMIT_PER_URL_GROUP, ROOT, REALTIME_UPDATE_INTERVAL


class UWSGILogParser(object):
    """Parser for uwsgi log file, support only default log format:

    log format: "[pid: 27011|app: 0|req: 16858/537445] 58.251.73.227 () {40 vars in 1030 bytes} \
                 [Tue Apr 29 00:13:10 2014] POST /trips/2387949771/add_waypoint/ => \
                 generated 1053 bytes in 2767 msecs (HTTP/1.1 200) 4 headers in 282 bytes \
                 (1 switches on core 0)"

    Returns:
    ~~~~~~~~

    An dict of parsed log result.
    """
    DATETIME_FORMAT = '%a %b %d %H:%M:%S %Y'
    RE_LOG_LINE = re.compile(r'''}\ \[(?P<datetime>.*?)\]\ (?P<request_method>POST|GET|DELETE|PUT|PATCH)\s
        (?P<request_uri>[^ ]*?)\ =>\ generated\ (?:.*?)\ in\ (?P<resp_msecs>\d+)\ msecs\s
        \(HTTP/[\d.]+\ (?P<resp_status>\d+)\)''', re.VERBOSE)

    def __init__(self):
        pass

    def parse(self, line):
        matched = self.RE_LOG_LINE.search(line)
        if matched:
            matched_dict = matched.groupdict()
            method = matched_dict['request_method']
            status = matched_dict['resp_status']
            if not method in FILTER_METHODS or status not in FILTER_STATUS:
                return

            url = matched_dict['request_uri'].replace('//', '/')
            url_path = url.split('?')[0]
            resp_time = int(matched_dict['resp_msecs'])
            request_datetime = datetime.datetime.strptime(matched_dict['datetime'], 
                                                          self.DATETIME_FORMAT)
            return {
                'method': method,
                'url': url,
                'url_path': url_path,
                'resp_time': resp_time,
                'status': status,
                'request_datetime': request_datetime
            }
        return


class URLClassifier(object):
    """A simple url classifier, current rules:
        
    - replacing sequential digits part by '(\d+)'    
    """

    RE_SIMPLIFY_URL = re.compile(r'(?<=/)\d+(/|$)')

    def __init__(self, user_defined_rules=[]):
        self.user_defined_rules = user_defined_rules

    def classify(self, url_path):
        """Classify an url"""
        for dict_api_url in self.user_defined_rules:
            api_url = dict_api_url['str']
            re_api_url = dict_api_url['re']
            if re_api_url.match(url_path[1:]):
                return api_url

        return self.RE_SIMPLIFY_URL.sub(r'(\\d+)/', url_path)


class LogAnalyzer(object):
    """Log analyzer"""

    def __init__(self, url_classifier=None, min_msecs=200, start_from_datetime=None):
        self.data = {}
        self.requests_counter = {'normal': 0, 'slow': 0}
        self.total_slow_duration = 0

        self.min_msecs = min_msecs
        self.start_from_datetime = start_from_datetime
        self.datetime_range = [None, None]

        self.url_classifier = url_classifier or URLClassifier()
        self.log_parser = UWSGILogParser()

    def analyze_line(self, line):
        line = line.strip()
        result = self.log_parser.parse(line)
        # Ignore invalid log
        if not result:
            return
        if self.start_from_datetime and result['request_datetime'] <= self.start_from_datetime:
            return

        self.requests_counter['normal'] += 1

        if not self.datetime_range[0]:
            self.datetime_range[0] = result['request_datetime']
        self.datetime_range[1] = result['request_datetime']
        if result['resp_time'] < self.min_msecs:
            return

        resp_time = result['resp_time']

        # Use url_classifier to classify url
        matched_url_rule = self.url_classifier.classify(result['url_path'])

        big_d = self.data.setdefault((result['method'], matched_url_rule), {
            'urls': {},
            'duration_agr_data': ValuesAggregation(),
        })

        big_d['duration_agr_data'].add_value(resp_time)
        big_d['urls'].setdefault(result['url'], ValuesAggregation()).add_value(resp_time)
        
        self.requests_counter['slow'] += 1
        self.total_slow_duration += resp_time

    def get_data(self):
        return {
            'requests_counter': self.requests_counter,
            'total_slow_duration': self.total_slow_duration,
            'datetime_range': self.datetime_range,
            'data_details': self.data
        }


class RealtimeLogAnalyzer(object):
    """Log analyzer for realtime support"""

    default_data = {
        'requests_counter': {'normal': 0, 'slow': 0},
        'total_slow_duration': 0,
        'data_details': {}
    }

    def __init__(self, url_classifier=None, min_msecs=200, start_from_datetime=None):
        self.data = {}
        self.min_msecs = min_msecs
        self.start_from_datetime = start_from_datetime
        self.last_analyzed_datetime = None

        self.url_classifier = url_classifier or URLClassifier()
        self.log_parser = UWSGILogParser()

    def analyze_line(self, line):
        line = line.strip()
        result = self.log_parser.parse(line)
        # Ignore invalid log
        if not result:
            return
        if self.start_from_datetime and result['request_datetime'] <= self.start_from_datetime:
            return
        request_datetime = result['request_datetime']
        self.last_analyzed_datetime = request_datetime
        groups = self.get_result_group_names(request_datetime)
        if not groups:
            return

        for group in groups:
            if group not in self.data:
                self.data[group] = copy.deepcopy(self.default_data)

        for group in groups:
            self.data[group]['requests_counter']['normal'] += 1
        if result['resp_time'] < self.min_msecs:
            return

        resp_time = result['resp_time']

        # Use url_classifier to classify url
        matched_url_rule = self.url_classifier.classify(result['url_path'])

        for group in groups:
            big_d = self.data[group]['data_details'].setdefault((result['method'], matched_url_rule), {
                'urls': {},
                'duration_agr_data': ValuesAggregation(),
            })

            big_d['duration_agr_data'].add_value(resp_time)
            big_d['urls'].setdefault(result['url'], ValuesAggregation()).add_value(resp_time)
            
            self.data[group]['requests_counter']['slow'] += 1
            self.data[group]['total_slow_duration'] += resp_time

    def get_result_group_names(self, request_datetime):
        """Only today/yesterday/last interval are valid datetime"""
        request_date = request_datetime.date()
        today = datetime.date.today()
        yesterday = datetime.date.today() - datetime.timedelta(days=1)
        result = []
        if total_seconds(datetime.datetime.now() - request_datetime) < REALTIME_UPDATE_INTERVAL:
            result.append('last_interval')
        if request_date == today:
            result.append(today.isoformat())
        elif request_date == yesterday:
            result.append(yesterday.isoformat())
        return result

    def get_data(self, key=None):
        if key:
            return self.data.get(key, self.default_data)
        return self.data

    def clean_data_by_key(self, key):
        try:
            del self.data[key]
        except KeyError:
            pass


def format_data(raw_data, limit_per_url_group=LIMIT_PER_URL_GROUP, limit_url_groups=LIMIT_URL_GROUPS):
    """Fomat data from LogAnalyzer for render purpose"""
    data = copy.deepcopy(raw_data)
    for k, v in list(data['data_details'].items()):
        # Only reserve first ``limit_per_url_group`` items
        v['urls'] = sorted(list(v['urls'].items()), key=lambda k_v: k_v[1].total,
                           reverse=True)[:limit_per_url_group]

    data_details = sorted(iter(data['data_details'].items()),
                          key=lambda k_v1: k_v1[1]["duration_agr_data"].total, 
                          reverse=True)[:limit_url_groups]

    if data['requests_counter']['normal']:
        slow_rate = format(data['requests_counter']['slow'] / \
                           float(data['requests_counter']['normal']), '.2%')
    else:
        slow_rate = '-'
    data.update({
        'slow_rate': slow_rate,
        'data_details': data_details,
    })
    return data





# -*- coding: utf-8 -*-
"""File tailer, Based on: https://github.com/six8/pytailer"""
import os
import re
import time
import logging

logger = logging.getLogger(__name__)


class NoNewLine(object):
    pass

no_new_line = NoNewLine()


class Tailer(object):
    """Implements tailing and heading functionality like GNU tail and head
    commands.
    """
    line_terminators = ('\r\n', '\n', '\r')
    DEFAULT_BLOCK_SIZE = 4096
    MAX_UNCHANGED_STATS = 5

    def __init__(self, file, read_size=DEFAULT_BLOCK_SIZE, end=False):
        if isinstance(file, str):
            file = open(file, 'r')

        self.should_stop_follow = False
        self.read_size = read_size
        self.file = file
        self.start_pos = self.file.tell()
        if end:
            self.seek_end()
    
    def splitlines(self, data):
        return re.split('|'.join(self.line_terminators), data)

    def seek_end(self):
        self.seek(0, 2)

    def seek(self, pos, whence=0):
        self.file.seek(pos, whence)

    def read(self, read_size=None):
        if read_size:
            read_str = self.file.read(read_size)
        else:
            read_str = self.file.read()

        return len(read_str), read_str

    def seek_line_forward(self):
        """\
        Searches forward from the current file position for a line terminator
        and seeks to the charachter after it.
        """
        pos = start_pos = self.file.tell()

        bytes_read, read_str = self.read(self.read_size)

        start = 0
        if bytes_read and read_str[0] in self.line_terminators:
            # The first charachter is a line terminator, don't count this one
            start += 1

        while bytes_read > 0:          
            # Scan forwards, counting the newlines in this bufferfull
            i = start
            while i < bytes_read:
                if read_str[i] in self.line_terminators:
                    self.seek(pos + i + 1)
                    return self.file.tell()
                i += 1

            pos += self.read_size
            self.seek(pos)

            bytes_read, read_str = self.read(self.read_size)

        return None

    def seek_line(self):
        """\
        Searches backwards from the current file position for a line terminator
        and seeks to the charachter after it.
        """
        pos = end_pos = self.file.tell()

        read_size = self.read_size
        if pos > read_size:
            pos -= read_size
        else:
            pos = 0
            read_size = end_pos

        self.seek(pos)

        bytes_read, read_str = self.read(read_size)

        if bytes_read and read_str[-1] in self.line_terminators:
            # The last charachter is a line terminator, don't count this one
            bytes_read -= 1

            if read_str[-2:] == '\r\n' and '\r\n' in self.line_terminators:
                # found crlf
                bytes_read -= 1

        while bytes_read > 0:          
            # Scan backward, counting the newlines in this bufferfull
            i = bytes_read - 1
            while i >= 0:
                if read_str[i] in self.line_terminators:
                    self.seek(pos + i + 1)
                    return self.file.tell()
                i -= 1

            if pos == 0 or pos - self.read_size < 0:
                # Not enought lines in the buffer, send the whole file
                self.seek(0)
                return None

            pos -= self.read_size
            self.seek(pos)

            bytes_read, read_str = self.read(self.read_size)

        return None
  
    def tail(self, lines=10):
        """\
        Return the last lines of the file.
        """
        self.seek_end()
        end_pos = self.file.tell()

        for i in range(lines):
            if not self.seek_line():
                break

        data = self.file.read(end_pos - self.file.tell() - 1)
        if data:
            return self.splitlines(data)
        else:
            return []
               
    def head(self, lines=10):
        """\
        Return the top lines of the file.
        """
        self.seek(0)

        for i in range(lines):
            if not self.seek_line_forward():
                break
    
        end_pos = self.file.tell()
        
        self.seek(0)
        data = self.file.read(end_pos - 1)

        if data:
            return self.splitlines(data)
        else:
            return []

    def follow(self, delay=1.0):
        """\
        Iterator generator that returns lines as data is added to the file.

        Based on: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/157035
        """
        # TODO: Handle log file rotation
        self.trailing = True
        unchanged_stats = 0
        
        while not self.should_stop_follow:
            where = self.file.tell()
            line = self.file.readline()
            if line:    
                if self.trailing and line in self.line_terminators:
                    # This is just the line terminator added to the end of the file
                    # before a new line, ignore.
                    self.trailing = False
                    continue

                if line[-1] in self.line_terminators:
                    line = line[:-1]
                    if line[-1:] == '\r\n' and '\r\n' in self.line_terminators:
                        # found crlf
                        line = line[:-1]

                self.trailing = False
                unchanged_stats = 0
                yield line
            else:
                self.trailing = True
                self.seek(where)
                yield no_new_line
                # Try to catch up rotated log file
                unchanged_stats += 1
                if unchanged_stats >= self.MAX_UNCHANGED_STATS and \
                        where != os.stat(self.file.name).st_size:
                    logger.info('Reopen log file because file may has been rotated.')
                    self.reopen_file()

                time.sleep(delay)

    def reopen_file(self):
        self.file = open(self.file.name, 'r')

    def stop_follow(self):
        self.should_stop_follow = True

    def __iter__(self):
        return self.follow()

    def close(self):
        self.file.close()





# -*- coding: utf-8 -*-
"""Start uwsgi-sloth workers"""
import os
import signal
import argparse
import datetime
from configobj import ConfigObj

from uwsgi_sloth.analyzer import format_data, RealtimeLogAnalyzer, URLClassifier
from uwsgi_sloth.tailer import Tailer, no_new_line
from uwsgi_sloth.template import render_template
from uwsgi_sloth.utils import makedir_if_none_exists, total_seconds, parse_url_rules
from uwsgi_sloth.models import merge_requests_data_to, RequestsData, SavePoint
from uwsgi_sloth.settings import REALTIME_UPDATE_INTERVAL, DEFAULT_MIN_MSECS

import logging
logger = logging.getLogger('uwsgi_sloth')


class HTMLRender(object):
    """helper for render HTML"""
    def __init__(self, html_dir, domain=None):
        self.html_dir = html_dir
        self.domain = domain

    def render_requests_data_to_html(self, data, file_name, context={}):
        """Render to HTML file"""
        file_path = os.path.join(self.html_dir, file_name)
        logger.info('Rendering HTML file %s...' % file_path)
        data = format_data(data)
        data.update(context)
        data.update(domain=self.domain)
        with open(file_path, 'w') as fp:
            fp.write(render_template('realtime.html', data))


def update_html_symlink(html_dir):
    """"Maintail symlink: "today.html", "yesterday.html" """
    today = datetime.date.today()
    yesterday = datetime.date.today() - datetime.timedelta(days=1)

    for from_date, alias_name in (
            (today, 'today.html'), (yesterday, 'yesterday.html')):
        from_date_file_path = os.path.join(html_dir, 'day_%s.html' % from_date)
        symlink_path = os.path.join(html_dir, alias_name)
        try:                                                                                         
            os.unlink(symlink_path)                                                                  
        except OSError:                                                                              
            pass                                                                                     
        os.symlink(from_date_file_path, symlink_path)


def start(args):
    # Load config file
    config = ConfigObj(infile=args.config.name)
    data_dir = config['data_dir']
    uwsgi_log_path = config['uwsgi_log_path']
    min_msecs = int(config.get('min_msecs', DEFAULT_MIN_MSECS))
    url_file = config.get('url_file')

    # Load custom url rules
    url_rules = []
    if url_file:
        with open(url_file, 'r') as fp:
            url_rules = parse_url_rules(fp)

    html_dir = os.path.join(data_dir, 'html')
    db_dir = os.path.join(data_dir, 'data')
    makedir_if_none_exists(html_dir)
    makedir_if_none_exists(db_dir)

    save_point = SavePoint(db_dir)
    last_log_datetime = save_point.get_last_datetime() or \
                    (datetime.datetime.now() - datetime.timedelta(seconds=REALTIME_UPDATE_INTERVAL))
    logger.info('Start from last savepoint, last_log_datetime: %s' % last_log_datetime)

    last_update_datetime = None
    url_classifier = URLClassifier(user_defined_rules=url_rules)
    analyzer = RealtimeLogAnalyzer(url_classifier=url_classifier, min_msecs=min_msecs,
                                   start_from_datetime=last_log_datetime)
    file_tailer = Tailer(uwsgi_log_path)
    html_render = HTMLRender(html_dir, domain=config.get('domain'))

    # Listen INT/TERM signal
    def gracefully_exit(*args):
        logger.info('Sinal received, exit.')
        file_tailer.stop_follow()
    signal.signal(signal.SIGINT, gracefully_exit) 

    for line in file_tailer:
        # Analyze line
        if line != no_new_line:
            analyzer.analyze_line(line)

        now = datetime.datetime.now()
        if not file_tailer.trailing:
            continue
        if last_update_datetime and \
                total_seconds(now - last_update_datetime) < REALTIME_UPDATE_INTERVAL:
            continue

        # Render HTML file when:
        # - file_tailer reaches end of file.
        # - last_update_datetime if over one `interval` from now

        # Render latest interval HTML file
        html_render.render_requests_data_to_html(analyzer.get_data('last_interval'),
             'latest_5mins.html', context={'datetime_range': 'Last 5 minutes'})
        analyzer.clean_data_by_key('last_interval')

        for date in list(analyzer.data.keys()):
            day_requests_data = RequestsData(date, db_dir)
            merge_requests_data_to(day_requests_data.data, analyzer.get_data(date))
            # Render to HTML file
            html_render.render_requests_data_to_html(day_requests_data.data,
                'day_%s.html' % date, context={'datetime_range': date})
            # Save data to pickle file
            day_requests_data.save()
            # Reset Everything
            analyzer.clean_data_by_key(date)

        update_html_symlink(html_dir)
        last_update_datetime = now
        if analyzer.last_analyzed_datetime:
            save_point.set_last_datetime(analyzer.last_analyzed_datetime)
            save_point.save()


def load_subcommand(subparsers):
    """Load this subcommand"""
    parser_start = subparsers.add_parser('start', help='Start uwsgi-sloth process for realtime analyzing.')
    parser_start.add_argument('-c', '--config', type=argparse.FileType('r'), dest='config',
                                help='uwsgi-sloth config file, use "uwsgi-sloth echo_conf" for a default one', required=True)
    parser_start.set_defaults(func=start)




# -*- coding: utf-8 -*-
import sys
import time
import logging
import argparse

from uwsgi_sloth.settings import LIMIT_URL_GROUPS, LIMIT_PER_URL_GROUP
from uwsgi_sloth.analyzer import URLClassifier, LogAnalyzer, format_data
from uwsgi_sloth.template import render_template
from uwsgi_sloth.utils import parse_url_rules

logger = logging.getLogger('uwsgi_sloth.analyze')


def analyze_log(fp, configs, url_rules):
    """Analyze log file"""
    url_classifier = URLClassifier(url_rules)
    analyzer = LogAnalyzer(url_classifier=url_classifier, min_msecs=configs.min_msecs)
    for line in fp:
        analyzer.analyze_line(line)
    return analyzer.get_data()


def analyze(args):
    # Get custom url rules
    url_rules = []
    if args.url_file:
        url_rules = parse_url_rules(args.url_file)

    logger.info('Analyzing log file "%s"...' % args.filepath.name)
    start_time = time.time()
    data = analyze_log(args.filepath, args, url_rules)
    data = format_data(data, args.limit_per_url_group, args.limit_url_groups)
    data.update({
        'domain': args.domain,
        'input_filename': args.filepath.name,
        'min_duration': args.min_msecs,
    })

    # Pre-process data
    html_data = render_template('report.html', data)

    args.output.write(html_data)
    args.output.close()
    logger.info('Finished in %.2f seconds.' % (time.time() - start_time))


def load_subcommand(subparsers):
    """Load this subcommand
    """
    parser_analyze = subparsers.add_parser('analyze', help='Analyze uwsgi log to get report')
    parser_analyze.add_argument('-f', '--filepath', type=argparse.FileType('r'), dest='filepath',
                                help='Path of uwsgi log file', required=True)
    parser_analyze.add_argument('--output', dest="output", type=argparse.FileType('w'), default=sys.stdout, 
                                help='HTML report file path')
    parser_analyze.add_argument('--min-msecs', dest="min_msecs", type=int, default=200,
                                help='Request serve time lower than this value will not be counted, default: 200')
    parser_analyze.add_argument('--domain', dest="domain", type=str, required=False,
                                help='Make url in report become a hyper-link by settings a domain')
    parser_analyze.add_argument('--url-file', dest="url_file", type=argparse.FileType('r'), required=False, 
                                help='Customized url rules in regular expression')
    parser_analyze.add_argument('--limit-url-groups', dest="limit_url_groups", type=int, required=False, 
                                default=LIMIT_URL_GROUPS, help='Number of url groups considered, default: 200')
    parser_analyze.add_argument('--limit-per-url-group', dest="limit_per_url_group", type=int,
                                required=False, default=LIMIT_PER_URL_GROUP,
                                help='Number of urls per group considered, default: 20')
    parser_analyze.set_defaults(func=analyze)



# -*- coding: utf-8 -*-
"""Useful data structures"""


class ValuesAggregation(object):
    """For response time analyze"""

    def __init__(self, values=[]):
        self.min = None
        self.max = None
        self.total = 0
        self.count = 0
        # Init with values
        for value in values:
            self.add_value(value)

    def add_value(self, value):
        self.count += 1
        self.total += value
        if self.max is None or value > self.max:
            self.max = value
        if self.min is None or value < self.min:
            self.min = value

    def add_values(self, values):
        for value in values:
            self.add_value(value)

    @property
    def avg(self):
        if not self.count:
            return 0
        return self.total / float(self.count)

    def merge_with(self, other):
        """Merge this ``ValuesAggregation`` with another one"""
        result = ValuesAggregation()
        result.total = self.total + other.total
        result.count = self.count + other.count
        result.min = min(self.min, other.min)
        result.max = max(self.max, other.max)
        return result

    def get_result(self):
        return {
            'min': self.min,
            'max': self.max,
            'avg': self.avg,
        }




# -*- coding: utf-8 -*-
from __future__ import print_function

import sys
from setuptools import setup, find_packages


# **Python version check**
if sys.version_info < (3, 5):
    error = """
uwsgi-sloth only supports Python 3.5 and above.
If you are using Python 2.7, please install "uwsgi-sloth<3.0.0" instead.
"""

    print(error, file=sys.stderr)
    sys.exit(1)


setup(
    name='uwsgi-sloth',
    version='3.0.2',
    description='A simple uwsgi access log analyzer',
    long_description=open('README.rst').read(),
    author='piglei',
    author_email='piglei2007@gmail.com',
    url='https://github.com/piglei/uwsgi-sloth',
    keywords='uwsgi log analyzer',
    license='Apache License, Version 2.0',
    packages=find_packages(),
    package_data={"": ['templates/*.html', 'sample.conf']},
    classifiers=[
      "Programming Language :: Python :: 3",
    ],
    install_requires=[
        'jinja2',
        'configobj'
    ],
    scripts=['uwsgi_sloth/uwsgi-sloth']
)



# -*- coding: utf-8 -*-

__version__ = '3.0.2'
__VERSION__ = __version__

# Init logger
import logging
import logging.handlers
logger = logging.getLogger('uwsgi_sloth')
logger.setLevel(logging.DEBUG)

handler = logging.StreamHandler()
formatter = logging.Formatter('[%(asctime)s] %(name)s %(levelname)s: %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)

handler.setLevel(logging.DEBUG)

import uwsgi_sloth.compat



# -*- coding: utf-8 -*-
"""Test code for ananlyzer"""
from uwsgi_sloth.analyzer import UWSGILogParser, URLClassifier


class TestUWSGILogParser(object):

    @classmethod
    def setup_class(cls):
        cls.parser = UWSGILogParser()
        cls.invalid_log_line = 'INVALID LOG LINE'
        cls.valid_log_line = ('[pid: 94153|app: 0|req: 51/51] 127.0.0.1 () {40 vars in 880 bytes} '
                              '[Tue Jun 24 22:46:02 2014] GET /trips/hot/?query=3 => generated '
                              '16432 bytes in 55 msecs (HTTP/1.1 200) 4 headers in 285 bytes '
                              '(4 switches on core 0)')

    def test_parse(self):
        valid_result = self.parser.parse(self.valid_log_line)
        assert valid_result['status'] == '200'
        assert valid_result['url'] == '/trips/hot/?query=3'
        assert valid_result['url_path'] == '/trips/hot/'
        assert valid_result['resp_time'] == 55


class TestURLClassifier(object):
    @classmethod
    def setup_class(cls):
        cls.url_classifier = URLClassifier()

    def test_classify_digits(self):
        path = self.url_classifier.classify('/trips/hot/42/foo')
        assert path == '/trips/hot/(\d+)/foo'

    def test_classify_trailing_digits(self):
        path = self.url_classifier.classify('/trips/hot/42')
        assert path == '/trips/hot/(\d+)/'



# -*- coding: utf-8 -*-
import os
import re

def parse_url_rules(urls_fp):
    """URL rules from given fp"""
    url_rules = []
    for line in urls_fp:
        re_url = line.strip()
        if re_url:
            url_rules.append({'str': re_url, 're': re.compile(re_url)})
    return url_rules


def makedir_if_none_exists(d):
    if not os.path.exists(d):
        os.makedirs(d)


def total_seconds(td):
    """Return timedelta's total seconds"""
    return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 1e6) / 1e6


def force_bytes(s, encoding='utf-8', errors='strict'):
    """A function turns "s" into bytes object, similar to django.utils.encoding.force_bytes
    """
    # Handle the common case first for performance reasons.
    if isinstance(s, bytes):
        if encoding == 'utf-8':
            return s
        else:
            return s.decode('utf-8', errors).encode(encoding, errors)
    else:
        return s.encode(encoding, errors)

    
def force_text(s, encoding='utf-8',  errors='strict'):
    """A function turns "s" into text type, similar to django.utils.encoding.force_text
    """
    if issubclass(type(s), str):
        return s
    try:
        if isinstance(s, bytes):
            s = str(s, encoding, errors)
        else:
            s = str(s)
    except UnicodeDecodeError as e:
        raise DjangoUnicodeDecodeError(s, *e.args)
    return s